{$Stackframes on}

{.Define HandleRunerrors}
{ Define this if you want a writeln-style runtime error handler
  to be installed.  Only useful in console mode programs, so
  not defined by default. }

unit DebugLow;

interface

uses
  windows,classes;

function __file__:shortstring;
{ Gives the filename of the source code line where the
  call occurs. }

function __line__:word;
{ Gives the line number of the source code line where the
  call occurs }

function SrcLine(addr:pointer):string;
{ Gives the source file and line number of addr, or
  blank if not found }

function Procname(addr:pointer):string;
{ Gives the unit and procedure name of containing addr, or
  an approximation to them }

function Dataname(addr:pointer):string;
{ Gives the unit and name for global data addresses, or an
  approximation to them }

function StackTrace(Skip:integer):TStringlist;
{ Walks up the stack frame records, reporting on the
  procedure and source line as it goes.  Skips back Skip
  stack frames before it starts }

function StackDump(Skip,maxentries:integer;full:boolean):TStringList;
{ Looks upwards through the stack for recognizable addresses.
  They won't all be valid! Skips back Skip stack frames
  before it starts.  If full is true, then unrecognized values
  are shown too }

procedure Initialize(filename:string);
{ Set the EXE file to a new file. }

function StackLimit:integer;
{ The address just beyond the top (logical bottom) of the stack.
  All addresses on the stack should be between ESP and StackLimit }

function CallerAddr:pointer;
{ The address of the caller of the current function. Requires
  stack frames }

function CurrentIP:pointer;
{ The current address }

function PtrToHex(p:pointer):string;

var
  ExeName : string;
const
  DJMDebugError : string = 'DJMDebug not initialized';

implementation

uses
  sysutils;

Type
  TInitDebug = function (filename:shortstring):integer; stdcall;
  TCloseDebug = procedure (handle:integer); stdcall;
  TGetLineInfo = procedure (handle:integer;addr:pointer;var filename:shortstring;var linenum:word); stdcall;
  TGetProcInfo = procedure (handle:integer;addr:pointer;var modulename,procname:shortstring); stdcall;
  TGetDataInfo = procedure (handle:integer;addr:pointer;var modulename,dataname:shortstring); stdcall;

var
  InitDebug : TInitDebug;
  CloseDebug : TCloseDebug;
  GetLineInfo : TGetLineInfo;
  GetProcInfo : TGetProcInfo;
  GetDataInfo : TGetDataInfo;
  DJMDebugLoaded : boolean;
  hModule : longint;
  initialized : boolean;
  handle : integer;

procedure LoadDJMDebug;
begin
  if not DJMDebugLoaded then
  begin
    hModule := LoadLibrary('DJMDEBUG.DLL');
    DJMDebugLoaded := hModule <> 0;
    if DJMDebugLoaded then
    begin
      InitDebug := GetProcAddress(hModule,'InitDebug');
      CloseDebug := GetProcAddress(hModule,'CloseDebug');
      GetLineInfo := GetProcAddress(hModule,'GetLineInfo');
      GetProcInfo := GetProcAddress(hModule,'GetProcInfo');
      GetDataInfo := GetProcAddress(hModule,'GetDataInfo');
    end
    else
      DJMDebugError := 'DJMDEBUG.DLL not found';
  end;
end;

procedure UnloadDJMDebug;
begin
  if DJMDebugLoaded then
    DJMDebugLoaded := not FreeLibrary(hModule);
end;

function PtrToHex(p:pointer):string;
begin
  result := IntToHex(integer(p),8);
end;

procedure Initialize(filename:string);
begin
  LoadDJMDebug;
  if not DJMDebugLoaded then
    exit;
  if initialized then
  begin
    if filename = exename then
      exit;
    CloseDebug(handle);
  end;
  handle := InitDebug(filename);
  initialized := handle <> 0;
  if initialized then
    DJMDebugError := ''
  else
    DJMDebugError := 'TD32 info not found in '+filename;
  ExeName := filename;
end;

function CallerAddr:pointer;
assembler;
asm
  mov eax,[ebp+4]
  sub eax,5
end;

function CurrentIP:pointer;
assembler;
asm
  mov eax,[esp]
  sub eax,5
end;

function EBPtoCaller(ebp:longint):pointer;
assembler;
asm
  mov eax,[eax+4]
end;

function PrevBP(ebp:longint):longint;
assembler;
asm
  mov eax,[eax]
end;

function __Line__:word;
var
  filename : shortstring;
begin
  GetLineInfo(Handle,CallerAddr,filename,result);
end;

function __File__:shortstring;
var
  linenum : word;
begin
  if not initialized then
    result := DJMDebugError
  else
    GetLineInfo(Handle,CallerAddr,result,linenum);
end;

function Srcline(addr:pointer):string;
var
  filename:shortstring;
  linenum : word;
begin
  result := '';
  if initialized then
  begin
    GetLineInfo(Handle,addr,filename,linenum);
    if linenum <> 0 then
      result := filename+'#'+IntToStr(linenum);
  end;
end;

function Procname(addr:pointer):string;
var
  modname,pname : shortstring;
begin
  if not initialized then
    result := DJMDebugError
  else
  begin
    GetProcInfo(Handle,Addr,modname,pname);
    result := modname+'.'+pname;
  end;
end;

function Dataname(addr:pointer):string;
var
  modname,dname : shortstring;
begin
  if not initialized then
    result := DJMDebugError
  else
  begin
    GetDataInfo(Handle,Addr,modname,dname);
    result := modname+'.'+dname;
  end;
end;

function StackLimit:longint;
var
  si : TSystemInfo;
  mbi : TMemoryBasicInformation;
  pagesize : integer;
  mysp : integer;
begin
  GetSystemInfo(si);
  pagesize := si.dwPageSize;
  asm
    mov mysp,esp
  end;
  mysp := pagesize*(mysp div pagesize);
  VirtualQuery(pointer(mysp),mbi,sizeof(mbi));
  result := mbi.RegionSize + integer(mbi.BaseAddress);
end;

function GetEBP(skip:integer):integer;
assembler;
asm
  mov ecx,skip
  mov eax,ebp
  jecxz @2
@1:
  mov eax,[eax]
  loop @1
@2:
end;

function StackTrace(skip:integer):TStringList;
var
  myebp : integer;
  addr : pchar;
  proc,src : string;
begin
  myebp := GetEBP(skip);
  result := TStringList.Create;
  while myebp < Stacklimit-4 do
  begin
    try
      addr := EBPtoCaller(myebp);
    except
      on EAccessViolation do
        break;
    end;
    proc := Procname(addr-1);
    src := SrcLine(addr-1);
    if src <> '' then
      proc := proc + ' ['+src+']';
    result.add(IntToHex(myebp+4,8)+':  '
               +PtrToHex(addr)+' RETURN '+proc);
    if PrevBP(myebp) - myebp <= 0 then
      break;
    myebp := PrevBP(myebp);
  end;
end;

type
  char4 = array[1..4] of char;

function Nice(s:char4):char4;
var
  i : integer;
begin
  result := s;
  for i:=1 to 4 do
    if not (result[i] in [#32..#126,#128..#255]) then
      result[i] := '.';
end;
      
function StackDump(skip,maxentries:integer;full:boolean):TStringList;
var
  mysp : pointer;
  addr : pchar;
  ident,src : string;
  limit : integer;
begin
  limit := StackLimit;
  mysp := Pointer(GetEBP(skip));
  result := TStringList.Create;
  repeat
    try
      addr := pointer(mysp^);
    except
      on Exception do
        break;
    end;
    ident := procname(addr-1);
    if not (ident[1] in ['.','?']) then
    begin
      src := SrcLine(addr-1);
      if src <> '' then
        ident := ident + '[' + src + ']';
      result.add(PtrToHex(mysp)+': '+
                 PtrToHex(addr)+' PROC/FN '+ident);
    end
    else
    begin
      ident := dataname(addr);
      if ident[1] <> '?' then
        result.add(PtrToHex(mysp)+': '+PtrToHex(addr)+' DATA    '+ident)
      else if full then
        result.add(PtrToHex(mysp)+': '+PtrToHex(addr)+'         '+nice(Char4(addr)));
    end;
    mysp := Pointer(Integer(mysp)+4);
  until (result.Count >= maxentries) or (integer(mysp) >= limit);
end;

{$ifdef handlerunerrors}
procedure MyExitProc;
begin
  if assigned(erroraddr) then
  begin
    { Initialize just in case }
    Initialize(paramstr(0));
    writeln('Run time error ',exitcode,' at ',PtrToHex(ErrorAddr),' in ',
            Procname(Pchar(ErrorAddr)-1),' at line ',SrcLine(pchar(ErrorAddr)-1));
    writeln('Hit enter...');            
    readln;            
    erroraddr := nil;
  end;
end;
{$endif}

initialization
{$ifdef handlerunerrors}
  AddExitProc(MyExitProc);
{$endif}
  DJMDebugLoaded := false;
finalization
  if initialized then
   CloseDebug(handle);
  if DJMDebugLoaded then
    UnloadDJMDebug;
end.
